# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""File IO methods that wrap the C++ FileSystem API.

The C++ FileSystem API is SWIG wrapped in file_io.i. These functions call those
to accomplish basic File IO operations.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import binascii
import collections
import glob
import os
import shutil
import six
import uuid

from . import compat, errors


# A good default block size depends on the system in question.
# A somewhat conservative default chosen here.
_DEFAULT_BLOCK_SIZE = 16 * 1024 * 1024


class FileIO(object):
    # Only methods needed for TensorBoard are implemented.

    def __init__(self, filename, mode):
        self.filename = compat.as_bytes(filename)
        self.mode = compat.as_bytes(mode)
        self.f = open(self.filename, self.mode)

    def __del__(self):
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.f.close()

    def __iter__(self):
        return self.f

    def size(self):
        return os.stat(self.filename).st_size

    def read(self):
        return self.f.read()

    def close(self):
        self.f.close()


class GFile(FileIO):
    # Same interface as FileIO but through GFile class

    def __init__(self, filename, mode):
        super(GFile, self).__init__(filename, mode)


# @tf_export("gfile.Exists")
def Exists(filename):
    """Determines whether a path exists or not.

    Args:
      filename: string, a path

    Returns:
      True if the path exists, whether its a file or a directory.
      False if the path does not exist and there are no filesystem errors.

    Raises:
      errors.OpError: Propagates any errors reported by the FileSystem API.
    """
    try:
        return os.path.exists(compat.as_bytes(filename))
    except errors.NotFoundError:
        return False
    return True


# @tf_export("gfile.Remove")
def Remove(filename):
    """Deletes the file located at 'filename'.

    Args:
        filename: string, a filename

    Raises:
        errors.OpError: Propagates any errors reported by the FileSystem API.
        E.g., NotFoundError if the file does not exist.
    """
    fn = compat.as_bytes(filename)
    if not os.path.exists(fn) or not os.path.isfile(fn):
        return
    os.remove(fn)


def read_file_to_string(filename, binary_mode=False):
    """Reads the entire contents of a file to a string.

    Args:
      filename: string, path to a file
      binary_mode: whether to open the file in binary mode or not. This changes
          the type of the object returned.

    Returns:
      contents of the file as a string or bytes.

    Raises:
      errors.OpError: Raises variety of errors that are subtypes e.g.
      NotFoundError etc.
    """
    if binary_mode:
        f = FileIO(filename, mode="rb")
    else:
        f = FileIO(filename, mode="r")
    return f.read()


def write_string_to_file(filename, file_content):
    """Writes a string to a given file.

    Args:
      filename: string, path to a file
      file_content: string, contents that need to be written to the file

    Raises:
      errors.OpError: If there are errors during the operation.
    """
    with FileIO(filename, mode="w") as f:
        f.write(file_content)


# @tf_export("gfile.Glob")
def Glob(filename):
    """Returns a list of files that match the given pattern(s).

    Args:
    filename: string or iterable of strings. The glob pattern(s).

    Returns:
    A list of strings containing filenames that match the given pattern(s).

    Raises:
    errors.OpError: If there are filesystem / directory listing errors.
    """
    if isinstance(filename, six.string_types):
        return [
            # compat the filenames to string from bytes.
            compat.as_str_any(matching_filename)
            for matching_filename in glob.glob(compat.as_bytes(filename))
        ]
    else:
        return [
            # compat the filenames to string from bytes.
            compat.as_str_any(matching_filename)
            for single_filename in filename
            for matching_filename in glob.glob(compat.as_bytes(single_filename))
        ]


# @tf_export("gfile.MkDir")
def MkDir(dirname):
    """Creates a directory with the name 'dirname'.

    Args:
    dirname: string, name of the directory to be created

    Notes:
    The parent directories need to exist. Use recursive_create_dir instead if
    there is the possibility that the parent dirs don't exist.

    Raises:
    errors.OpError: If the operation fails.
    """
    os.mkdir(compat.as_bytes(dirname))


# @tf_export("gfile.MakeDirs")
def MakeDirs(dirname):
    """Creates a directory and all parent/intermediate directories.

    It succeeds if dirname already exists and is writable.

    Args:
    dirname: string, name of the directory to be created

    Raises:
    errors.OpError: If the operation fails.
    """
    os.makedirs(compat.as_bytes(dirname))


# @tf_export("gfile.Copy")
def Copy(oldpath, newpath, overwrite=False):
    """Copies data from oldpath to newpath.

    Args:
    oldpath: string, name of the file who's contents need to be copied
    newpath: string, name of the file to which to copy to
    overwrite: boolean, if false its an error for newpath to be occupied by an
        existing file.

    Raises:
    errors.OpError: If the operation fails.
    """
    newpath_exists = os.path.exists(newpath)
    if newpath_exists and overwrite or not newpath_exists:
        shutil.copy2(compat.as_bytes(oldpath), compat.as_bytes(newpath))


# @tf_export("gfile.Rename")
def Rename(oldname, newname, overwrite=False):
    """Rename or move a file / directory.

    Args:
      oldname: string, pathname for a file
      newname: string, pathname to which the file needs to be moved
      overwrite: boolean, if false it's an error for `newname` to be occupied by
          an existing file.

    Raises:
      errors.OpError: If the operation fails.
    """
    newname_exists = os.path.exists(newname)
    if newname_exists and overwrite or not newname_exists:
        os.rename(compat.as_bytes(oldname), compat.as_bytes(newname))


def atomic_write_string_to_file(filename, contents, overwrite=True):
    """Writes to `filename` atomically.

    This means that when `filename` appears in the filesystem, it will contain
    all of `contents`. With write_string_to_file, it is possible for the file
    to appear in the filesystem with `contents` only partially written.

    Accomplished by writing to a temp file and then renaming it.

    Args:
      filename: string, pathname for a file
      contents: string, contents that need to be written to the file
      overwrite: boolean, if false it's an error for `filename` to be occupied by
          an existing file.
    """
    temp_pathname = filename + ".tmp" + uuid.uuid4().hex
    write_string_to_file(temp_pathname, contents)
    try:
        Rename(temp_pathname, filename, overwrite)
    except Exception:
        Remove(temp_pathname)
        raise


# @tf_export("gfile.DeleteRecursively")
def DeleteRecursively(dirname):
    """Deletes everything under dirname recursively.

    Args:
      dirname: string, a path to a directory

    Raises:
      errors.OpError: If the operation fails.
    """
    os.removedirs(dirname)


# @tf_export("gfile.IsDirectory")
def IsDirectory(dirname):
    """Returns whether the path is a directory or not.

    Args:
      dirname: string, path to a potential directory

    Returns:
      True, if the path is a directory; False otherwise
    """
    return os.path.isdir(compat.as_bytes(dirname))


# @tf_export("gfile.ListDirectory")
def ListDirectory(dirname):
    """Returns a list of entries contained within a directory.

    The list is in arbitrary order. It does not contain the special entries "."
    and "..".

    Args:
      dirname: string, path to a directory

    Returns:
      [filename1, filename2, ... filenameN] as strings

    Raises:
      errors.NotFoundError if directory doesn't exist
    """
    if not IsDirectory(dirname):
        raise errors.NotFoundError(None, None, "Could not find directory")
    try:
        entries = os.listdir(compat.as_bytes(dirname))

    except Exception:
        entries = os.listdir(compat.as_str_any(dirname))

    entries = [compat.as_str_any(item) for item in entries]
    return entries


# @tf_export("gfile.Walk")
def Walk(top, in_order=True):
    """Recursive directory tree generator for directories.

    Args:
      top: string, a Directory name
      in_order: bool, Traverse in order if True, post order if False.

    Errors that happen while listing directories are ignored.

    Yields:
      Each yield is a 3-tuple:  the pathname of a directory, followed by lists of
      all its subdirectories and leaf files.
      (dirname, [subdirname, subdirname, ...], [filename, filename, ...])
      as strings
    """
    top = compat.as_str_any(top)
    try:
        listing = ListDirectory(top)
    except errors.NotFoundError:
        return

    files = []
    subdirs = []
    for item in listing:
        full_path = os.path.join(top, item)
        if IsDirectory(full_path):
            subdirs.append(item)
        else:
            files.append(item)

    here = (top, subdirs, files)

    if in_order:
        yield here

    for subdir in subdirs:
        for subitem in os.walk(os.path.join(top, subdir), in_order):
            yield subitem

    if not in_order:
        yield here


# Data returned from the Stat call.
StatData = collections.namedtuple('StatData', ['length'])


# @tf_export("gfile.Stat")
def Stat(filename):
    """Returns file statistics for a given path.

    Args:
      filename: string, path to a file

    Returns:
      FileStatistics struct that contains information about the path

    Raises:
      errors.OpError: If the operation fails.
    """
    result = None
    try:
        # Size of the file is given by .st_size as returned from
        # os.stat() but TB uses .length so we set length.
        len = os.stat(compat.as_bytes(filename)).st_size
        result = StatData(len)
    except Exception:
        pass

    if result is None:
        raise errors.NotFoundError(None, None, 'Unable to stat file')
    return result


def filecmp(filename_a, filename_b):
    """Compare two files, returning True if they are the same, False otherwise.

    We check size first and return False quickly if the files are different sizes.
    If they are the same size, we continue to generating a crc for the whole file.

    You might wonder: why not use Python's filecmp.cmp() instead? The answer is
    that the builtin library is not robust to the many different filesystems
    TensorFlow runs on, and so we here perform a similar comparison with
    the more robust FileIO.

    Args:
      filename_a: string path to the first file.
      filename_b: string path to the second file.

    Returns:
      True if the files are the same, False otherwise.
    """
    size_a = FileIO(filename_a, "rb").size()
    size_b = FileIO(filename_b, "rb").size()
    if size_a != size_b:
        return False

    # Size is the same. Do a full check.
    crc_a = file_crc32(filename_a)
    crc_b = file_crc32(filename_b)
    return crc_a == crc_b


def file_crc32(filename, block_size=_DEFAULT_BLOCK_SIZE):
    """Get the crc32 of the passed file.

    The crc32 of a file can be used for error checking; two files with the same
    crc32 are considered equivalent. Note that the entire file must be read
    to produce the crc32.

    Args:
      filename: string, path to a file
      block_size: Integer, process the files by reading blocks of `block_size`
        bytes. Use -1 to read the file as once.

    Returns:
      hexadecimal as string, the crc32 of the passed file.
    """
    crc = 0
    with FileIO(filename, mode="rb") as f:
        chunk = f.read(n=block_size)
        while chunk:
            crc = binascii.crc32(chunk, crc)
            chunk = f.read(n=block_size)
    return hex(crc & 0xFFFFFFFF)



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""TensorFlow component package for providing tf.summary from TensorBoard."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import importlib

import tensorflow as tf

# Re-export all symbols from the original tf.summary.
# pylint: disable=wildcard-import,unused-import,g-import-not-at-top

if getattr(tf, '__version__', '').startswith('2.'):
  from tensorflow.summary import *
else:
  try:
    # Check if we can directly import tf.compat.v2. We may not be able to if we
    # reached this import itself while importing tf.compat.v2. Use importlib
    # to simulate 'import tensorflow.compat.v2' but without binding local names
    # which are hard to clean up.
    #
    # Note that we can't use either of the following internally:
    #
    #   import tensorflow.compat.v2 as other
    #   from tensorflow.compat import v2
    #
    # The former raises AttributeError until python 3.7:
    # <https://bugs.python.org/issue30024>
    # The latter raises ImportError even when it shouldn't, until python 3.5:
    # <https://bugs.python.org/issue17636>
    #
    # In commemoration of this fiasco, I offer the following haiku:
    #
    #   import a module
    #   it works. but add "as other"
    #   AttributeError
    #
    importlib.import_module('tensorflow.compat.v2')
  except ImportError:
    # If that failed, go "under the hood" and attempt to directly import the
    # module that will become tf.compat.v2.summary.
    try:
      from tensorflow._api.v1.compat.v2.summary import *
    except ImportError:
      from tensorflow._api.v2.compat.v2.summary import *
  else:
    from tensorflow.compat.v2.summary import *

from tensorboard.summary.v2 import audio
from tensorboard.summary.v2 import histogram
from tensorboard.summary.v2 import image
from tensorboard.summary.v2 import scalar
from tensorboard.summary.v2 import text

del absolute_import, division, print_function, importlib, tf



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utilities for using TensorBoard in notebook contexts, like Colab.

These APIs are experimental and subject to change.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import datetime
import shlex
import sys

from tensorboard import manager


# Return values for `_get_context` (see that function's docs for
# details).
_CONTEXT_COLAB = "_CONTEXT_COLAB"
_CONTEXT_IPYTHON = "_CONTEXT_IPYTHON"
_CONTEXT_NONE = "_CONTEXT_NONE"


def _get_context():
  """Determine the most specific context that we're in.

  Returns:
    _CONTEXT_COLAB: If in Colab with an IPython notebook context.
    _CONTEXT_IPYTHON: If not in Colab, but we are in an IPython notebook
      context (e.g., from running `jupyter notebook` at the command
      line).
    _CONTEXT_NONE: Otherwise (e.g., by running a Python script at the
      command-line or using the `ipython` interactive shell).
  """
  # In Colab, the `google.colab` module is available, but the shell
  # returned by `IPython.get_ipython` does not have a `get_trait`
  # method.
  try:
    import google.colab
    import IPython
  except ImportError:
    pass
  else:
    if IPython.get_ipython() is not None:
      # We'll assume that we're in a Colab notebook context.
      return _CONTEXT_COLAB

  # In an IPython command line shell or Jupyter notebook, we can
  # directly query whether we're in a notebook context.
  try:
    import IPython
  except ImportError:
    pass
  else:
    ipython = IPython.get_ipython()
    if ipython is not None and ipython.has_trait("kernel"):
      return _CONTEXT_IPYTHON

  # Otherwise, we're not in a known notebook context.
  return _CONTEXT_NONE


def load_ipython_extension(ipython):
  """IPython API entry point.

  Only intended to be called by the IPython runtime.

  See:
    https://ipython.readthedocs.io/en/stable/config/extensions/index.html
  """
  _register_magics(ipython)


def _register_magics(ipython):
  """Register IPython line/cell magics.

  Args:
    ipython: An `InteractiveShell` instance.
  """
  ipython.register_magic_function(
      _start_magic,
      magic_kind="line",
      magic_name="tensorboard",
  )


def _start_magic(line):
  """Implementation of the `%tensorboard` line magic."""
  return start(line)


def start(args_string):
  """Launch and display a TensorBoard instance as if at the command line.

  Args:
    args_string: Command-line arguments to TensorBoard, to be
      interpreted by `shlex.split`: e.g., "--logdir ./logs --port 0".
      Shell metacharacters are not supported: e.g., "--logdir 2>&1" will
      point the logdir at the literal directory named "2>&1".
  """
  context = _get_context()
  try:
    import IPython
    import IPython.display
  except ImportError:
    IPython = None

  if context == _CONTEXT_NONE:
    handle = None
    print("Launching TensorBoard...")
  else:
    handle = IPython.display.display(
        IPython.display.Pretty("Launching TensorBoard..."),
        display_id=True,
    )

  def print_or_update(message):
    if handle is None:
      print(message)
    else:
      handle.update(IPython.display.Pretty(message))

  parsed_args = shlex.split(args_string, comments=True, posix=True)
  start_result = manager.start(parsed_args)

  if isinstance(start_result, manager.StartLaunched):
    _display(
        port=start_result.info.port,
        print_message=False,
        display_handle=handle,
    )

  elif isinstance(start_result, manager.StartReused):
    template = (
        "Reusing TensorBoard on port {port} (pid {pid}), started {delta} ago. "
        "(Use '!kill {pid}' to kill it.)"
    )
    message = template.format(
        port=start_result.info.port,
        pid=start_result.info.pid,
        delta=_time_delta_from_info(start_result.info),
    )
    print_or_update(message)
    _display(
        port=start_result.info.port,
        print_message=False,
        display_handle=None,
    )

  elif isinstance(start_result, manager.StartFailed):
    def format_stream(name, value):
      if value == "":
        return ""
      elif value is None:
        return "\n<could not read %s>" % name
      else:
        return "\nContents of %s:\n%s" % (name, value.strip())
    message = (
        "ERROR: Failed to launch TensorBoard (exited with %d).%s%s" %
        (
            start_result.exit_code,
            format_stream("stderr", start_result.stderr),
            format_stream("stdout", start_result.stdout),
        )
    )
    print_or_update(message)

  elif isinstance(start_result, manager.StartTimedOut):
    message = (
        "ERROR: Timed out waiting for TensorBoard to start. "
        "It may still be running as pid %d."
        % start_result.pid
    )
    print_or_update(message)

  else:
    raise TypeError(
        "Unexpected result from `manager.start`: %r.\n"
        "This is a TensorBoard bug; please report it."
        % start_result
    )


def _time_delta_from_info(info):
  """Format the elapsed time for the given TensorBoardInfo.

  Args:
    info: A TensorBoardInfo value.

  Returns:
    A human-readable string describing the time since the server
    described by `info` started: e.g., "2 days, 0:48:58".
  """
  now = datetime.datetime.now()
  then = info.start_time
  return str(now.replace(microsecond=0) - then.replace(microsecond=0))


def display(port=None, height=None):
  """Display a TensorBoard instance already running on this machine.

  Args:
    port: The port on which the TensorBoard server is listening, as an
      `int`, or `None` to automatically select the most recently
      launched TensorBoard.
    height: The height of the frame into which to render the TensorBoard
      UI, as an `int` number of pixels, or `None` to use a default value
      (currently 600).
  """
  _display(port=port, height=height, print_message=True, display_handle=None)


def _display(port=None, height=None, print_message=False, display_handle=None):
  """Internal version of `display`.

  Args:
    port: As with `display`.
    height: As with `display`.
    print_message: True to print which TensorBoard instance was selected
      for display (if applicable), or False otherwise.
    display_handle: If not None, an IPython display handle into which to
      render TensorBoard.
  """
  if height is None:
    height = 600

  if port is None:
    infos = manager.get_all()
    if not infos:
      raise ValueError("Can't display TensorBoard: no known instances running.")
    else:
      info = max(manager.get_all(), key=lambda x: x.start_time)
      port = info.port
  else:
    infos = [i for i in manager.get_all() if i.port == port]
    info = (
        max(infos, key=lambda x: x.start_time)
        if infos
        else None
    )

  if print_message:
    if info is not None:
      message = (
          "Selecting TensorBoard with {data_source} "
          "(started {delta} ago; port {port}, pid {pid})."
      ).format(
          data_source=manager.data_source_from_info(info),
          delta=_time_delta_from_info(info),
          port=info.port,
          pid=info.pid,
      )
      print(message)
    else:
      # The user explicitly provided a port, and we don't have any
      # additional information. There's nothing useful to say.
      pass

  fn = {
      _CONTEXT_COLAB: _display_colab,
      _CONTEXT_IPYTHON: _display_ipython,
      _CONTEXT_NONE: _display_cli,
  }[_get_context()]
  return fn(port=port, height=height, display_handle=display_handle)


def _display_colab(port, height, display_handle):
  """Display a TensorBoard instance in a Colab output frame.

  The Colab VM is not directly exposed to the network, so the Colab
  runtime provides a service worker tunnel to proxy requests from the
  end user's browser through to servers running on the Colab VM: the
  output frame may issue requests to https://localhost:<port> (HTTPS
  only), which will be forwarded to the specified port on the VM.

  It does not suffice to create an `iframe` and let the service worker
  redirect its traffic (`<iframe src="https://localhost:6006">`),
  because for security reasons service workers cannot intercept iframe
  traffic. Instead, we manually fetch the TensorBoard index page with an
  XHR in the output frame, and inject the raw HTML into `document.body`.

  By default, the TensorBoard web app requests resources against
  relative paths, like `./data/logdir`. Within the output frame, these
  requests must instead hit `https://localhost:<port>/data/logdir`. To
  redirect them, we change the document base URI, which transparently
  affects all requests (XHRs and resources alike).
  """
  import IPython.display
  shell = """
    <div id="root"></div>
    <script>
      (function() {
        window.TENSORBOARD_ENV = window.TENSORBOARD_ENV || {};
        window.TENSORBOARD_ENV["IN_COLAB"] = true;
        document.querySelector("base").href = "https://localhost:%PORT%";
        function fixUpTensorboard(root) {
          const tftb = root.querySelector("tf-tensorboard");
          // Disable the fragment manipulation behavior in Colab. Not
          // only is the behavior not useful (as the iframe's location
          // is not visible to the user), it causes TensorBoard's usage
          // of `window.replace` to navigate away from the page and to
          // the `localhost:<port>` URL specified by the base URI, which
          // in turn causes the frame to (likely) crash.
          tftb.removeAttribute("use-hash");
        }
        function executeAllScripts(root) {
          // When `script` elements are inserted into the DOM by
          // assigning to an element's `innerHTML`, the scripts are not
          // executed. Thus, we manually re-insert these scripts so that
          // TensorBoard can initialize itself.
          for (const script of root.querySelectorAll("script")) {
            const newScript = document.createElement("script");
            newScript.type = script.type;
            newScript.textContent = script.textContent;
            root.appendChild(newScript);
            script.remove();
          }
        }
        function setHeight(root, height) {
          // We set the height dynamically after the TensorBoard UI has
          // been initialized. This avoids an intermediate state in
          // which the container plus the UI become taller than the
          // final width and cause the Colab output frame to be
          // permanently resized, eventually leading to an empty
          // vertical gap below the TensorBoard UI. It's not clear
          // exactly what causes this problematic intermediate state,
          // but setting the height late seems to fix it.
          root.style.height = `${height}px`;
        }
        const root = document.getElementById("root");
        fetch(".")
          .then((x) => x.text())
          .then((html) => void (root.innerHTML = html))
          .then(() => fixUpTensorboard(root))
          .then(() => executeAllScripts(root))
          .then(() => setHeight(root, %HEIGHT%));
      })();
    </script>
  """.replace("%PORT%", "%d" % port).replace("%HEIGHT%", "%d" % height)
  html = IPython.display.HTML(shell)
  if display_handle:
    display_handle.update(html)
  else:
    IPython.display.display(html)


def _display_ipython(port, height, display_handle):
  import IPython.display
  iframe = IPython.display.IFrame(
      src="http://localhost:%d" % port,
      height=height,
      width="100%",
  )
  if display_handle:
    display_handle.update(iframe)
  else:
    IPython.display.display(iframe)


def _display_cli(port, height, display_handle):
  del height  # unused
  del display_handle  # unused
  message = "Please visit http://localhost:%d in a web browser." % port
  print(message)


def list():
  """Print a listing of known running TensorBoard instances.

  TensorBoard instances that were killed uncleanly (e.g., with SIGKILL
  or SIGQUIT) may appear in this list even if they are no longer
  running. Conversely, this list may be missing some entries if your
  operating system's temporary directory has been cleared since a
  still-running TensorBoard instance started.
  """
  infos = manager.get_all()
  if not infos:
    print("No known TensorBoard instances running.")
    return

  print("Known TensorBoard instances:")
  for info in infos:
    template = "  - port {port}: {data_source} (started {delta} ago; pid {pid})"
    print(template.format(
        port=info.port,
        data_source=manager.data_source_from_info(info),
        delta=_time_delta_from_info(info),
        pid=info.pid,
    ))



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""TensorBoard is a webapp for understanding TensorFlow runs and graphs.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import functools
import threading
import types


def lazy_load(name):
  """Decorator to define a function that lazily loads the module 'name'.

  This can be used to defer importing troublesome dependencies - e.g. ones that
  are large and infrequently used, or that cause a dependency cycle -
  until they are actually used.

  Args:
    name: the fully-qualified name of the module; typically the last segment
      of 'name' matches the name of the decorated function

  Returns:
    Decorator function that produces a lazy-loading module 'name' backed by the
    underlying decorated function.
  """
  def wrapper(load_fn):
    # Wrap load_fn to call it exactly once and update __dict__ afterwards to
    # make future lookups efficient (only failed lookups call __getattr__).
    @_memoize
    def load_once(self):
      if load_once.loading:
        raise ImportError("Circular import when resolving LazyModule %r" % name)
      load_once.loading = True
      try:
        module = load_fn()
      finally:
        load_once.loading = False
      self.__dict__.update(module.__dict__)
      load_once.loaded = True
      return module
    load_once.loading = False
    load_once.loaded = False

    # Define a module that proxies getattr() and dir() to the result of calling
    # load_once() the first time it's needed. The class is nested so we can close
    # over load_once() and avoid polluting the module's attrs with our own state.
    class LazyModule(types.ModuleType):
      def __getattr__(self, attr_name):
        return getattr(load_once(self), attr_name)

      def __dir__(self):
        return dir(load_once(self))

      def __repr__(self):
        if load_once.loaded:
          return '<%r via LazyModule (loaded)>' % load_once(self)
        return '<module %r via LazyModule (not yet loaded)>' % self.__name__

    return LazyModule(name)
  return wrapper


def _memoize(f):
  """Memoizing decorator for f, which must have exactly 1 hashable argument."""
  nothing = object()  # Unique "no value" sentinel object.
  cache = {}
  # Use a reentrant lock so that if f references the resulting wrapper we die
  # with recursion depth exceeded instead of deadlocking.
  lock = threading.RLock()
  @functools.wraps(f)
  def wrapper(arg):
    if cache.get(arg, nothing) == nothing:
      with lock:
        if cache.get(arg, nothing) == nothing:
          cache[arg] = f(arg)
    return cache[arg]
  return wrapper



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Information on the custom scalars plugin."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorboard.compat.proto import summary_pb2

# A special tag named used for the summary that stores the layout.
CONFIG_SUMMARY_TAG = 'custom_scalars__config__'

PLUGIN_NAME = 'custom_scalars'

def create_summary_metadata():
  """Create a `SummaryMetadata` proto for custom scalar plugin data.

  Returns:
    A `summary_pb2.SummaryMetadata` protobuf object.
  """
  return summary_pb2.SummaryMetadata(
      plugin_data=summary_pb2.SummaryMetadata.PluginData(
          plugin_name=PLUGIN_NAME))



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
"""A wrapper for TensorFlow SWIG-generated bindings."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import array
import os
import struct

from . import errors


TFE_DEVICE_PLACEMENT_WARN = 0
TFE_DEVICE_PLACEMENT_SILENT_FOR_INT32 = 0
TFE_DEVICE_PLACEMENT_SILENT = 0
TFE_DEVICE_PLACEMENT_EXPLICIT = 0


def __getattr__(attr):
    return 0


def TF_bfloat16_type():
    return 0


def masked_crc32c(data):
    x = u32(crc32c(data))
    return u32(((x >> 15) | u32(x << 17)) + 0xa282ead8)


def u32(x):
    return x & 0xffffffff


CRC_TABLE = (
    0x00000000, 0xf26b8303, 0xe13b70f7, 0x1350f3f4,
    0xc79a971f, 0x35f1141c, 0x26a1e7e8, 0xd4ca64eb,
    0x8ad958cf, 0x78b2dbcc, 0x6be22838, 0x9989ab3b,
    0x4d43cfd0, 0xbf284cd3, 0xac78bf27, 0x5e133c24,
    0x105ec76f, 0xe235446c, 0xf165b798, 0x030e349b,
    0xd7c45070, 0x25afd373, 0x36ff2087, 0xc494a384,
    0x9a879fa0, 0x68ec1ca3, 0x7bbcef57, 0x89d76c54,
    0x5d1d08bf, 0xaf768bbc, 0xbc267848, 0x4e4dfb4b,
    0x20bd8ede, 0xd2d60ddd, 0xc186fe29, 0x33ed7d2a,
    0xe72719c1, 0x154c9ac2, 0x061c6936, 0xf477ea35,
    0xaa64d611, 0x580f5512, 0x4b5fa6e6, 0xb93425e5,
    0x6dfe410e, 0x9f95c20d, 0x8cc531f9, 0x7eaeb2fa,
    0x30e349b1, 0xc288cab2, 0xd1d83946, 0x23b3ba45,
    0xf779deae, 0x05125dad, 0x1642ae59, 0xe4292d5a,
    0xba3a117e, 0x4851927d, 0x5b016189, 0xa96ae28a,
    0x7da08661, 0x8fcb0562, 0x9c9bf696, 0x6ef07595,
    0x417b1dbc, 0xb3109ebf, 0xa0406d4b, 0x522bee48,
    0x86e18aa3, 0x748a09a0, 0x67dafa54, 0x95b17957,
    0xcba24573, 0x39c9c670, 0x2a993584, 0xd8f2b687,
    0x0c38d26c, 0xfe53516f, 0xed03a29b, 0x1f682198,
    0x5125dad3, 0xa34e59d0, 0xb01eaa24, 0x42752927,
    0x96bf4dcc, 0x64d4cecf, 0x77843d3b, 0x85efbe38,
    0xdbfc821c, 0x2997011f, 0x3ac7f2eb, 0xc8ac71e8,
    0x1c661503, 0xee0d9600, 0xfd5d65f4, 0x0f36e6f7,
    0x61c69362, 0x93ad1061, 0x80fde395, 0x72966096,
    0xa65c047d, 0x5437877e, 0x4767748a, 0xb50cf789,
    0xeb1fcbad, 0x197448ae, 0x0a24bb5a, 0xf84f3859,
    0x2c855cb2, 0xdeeedfb1, 0xcdbe2c45, 0x3fd5af46,
    0x7198540d, 0x83f3d70e, 0x90a324fa, 0x62c8a7f9,
    0xb602c312, 0x44694011, 0x5739b3e5, 0xa55230e6,
    0xfb410cc2, 0x092a8fc1, 0x1a7a7c35, 0xe811ff36,
    0x3cdb9bdd, 0xceb018de, 0xdde0eb2a, 0x2f8b6829,
    0x82f63b78, 0x709db87b, 0x63cd4b8f, 0x91a6c88c,
    0x456cac67, 0xb7072f64, 0xa457dc90, 0x563c5f93,
    0x082f63b7, 0xfa44e0b4, 0xe9141340, 0x1b7f9043,
    0xcfb5f4a8, 0x3dde77ab, 0x2e8e845f, 0xdce5075c,
    0x92a8fc17, 0x60c37f14, 0x73938ce0, 0x81f80fe3,
    0x55326b08, 0xa759e80b, 0xb4091bff, 0x466298fc,
    0x1871a4d8, 0xea1a27db, 0xf94ad42f, 0x0b21572c,
    0xdfeb33c7, 0x2d80b0c4, 0x3ed04330, 0xccbbc033,
    0xa24bb5a6, 0x502036a5, 0x4370c551, 0xb11b4652,
    0x65d122b9, 0x97baa1ba, 0x84ea524e, 0x7681d14d,
    0x2892ed69, 0xdaf96e6a, 0xc9a99d9e, 0x3bc21e9d,
    0xef087a76, 0x1d63f975, 0x0e330a81, 0xfc588982,
    0xb21572c9, 0x407ef1ca, 0x532e023e, 0xa145813d,
    0x758fe5d6, 0x87e466d5, 0x94b49521, 0x66df1622,
    0x38cc2a06, 0xcaa7a905, 0xd9f75af1, 0x2b9cd9f2,
    0xff56bd19, 0x0d3d3e1a, 0x1e6dcdee, 0xec064eed,
    0xc38d26c4, 0x31e6a5c7, 0x22b65633, 0xd0ddd530,
    0x0417b1db, 0xf67c32d8, 0xe52cc12c, 0x1747422f,
    0x49547e0b, 0xbb3ffd08, 0xa86f0efc, 0x5a048dff,
    0x8ecee914, 0x7ca56a17, 0x6ff599e3, 0x9d9e1ae0,
    0xd3d3e1ab, 0x21b862a8, 0x32e8915c, 0xc083125f,
    0x144976b4, 0xe622f5b7, 0xf5720643, 0x07198540,
    0x590ab964, 0xab613a67, 0xb831c993, 0x4a5a4a90,
    0x9e902e7b, 0x6cfbad78, 0x7fab5e8c, 0x8dc0dd8f,
    0xe330a81a, 0x115b2b19, 0x020bd8ed, 0xf0605bee,
    0x24aa3f05, 0xd6c1bc06, 0xc5914ff2, 0x37faccf1,
    0x69e9f0d5, 0x9b8273d6, 0x88d28022, 0x7ab90321,
    0xae7367ca, 0x5c18e4c9, 0x4f48173d, 0xbd23943e,
    0xf36e6f75, 0x0105ec76, 0x12551f82, 0xe03e9c81,
    0x34f4f86a, 0xc69f7b69, 0xd5cf889d, 0x27a40b9e,
    0x79b737ba, 0x8bdcb4b9, 0x988c474d, 0x6ae7c44e,
    0xbe2da0a5, 0x4c4623a6, 0x5f16d052, 0xad7d5351,
)


CRC_INIT = 0

_MASK = 0xFFFFFFFF


def crc_update(crc, data):
    """Update CRC-32C checksum with data.
    Args:
      crc: 32-bit checksum to update as long.
      data: byte array, string or iterable over bytes.
    Returns:
      32-bit updated CRC-32C as long.
    """

    if type(data) != array.array or data.itemsize != 1:
        buf = array.array("B", data)
    else:
        buf = data

    crc ^= _MASK
    for b in buf:
        table_index = (crc ^ b) & 0xff
        crc = (CRC_TABLE[table_index] ^ (crc >> 8)) & _MASK
    return crc ^ _MASK


def crc_finalize(crc):
    """Finalize CRC-32C checksum.
    This function should be called as last step of crc calculation.
    Args:
      crc: 32-bit checksum as long.
    Returns:
      finalized 32-bit checksum as long
    """
    return crc & _MASK


def crc32c(data):
    """Compute CRC-32C checksum of the data.
    Args:
      data: byte array, string or iterable over bytes.
    Returns:
      32-bit CRC-32C checksum of data as long.
    """
    return crc_finalize(crc_update(CRC_INIT, data))


class PyRecordReader_New:
    def __init__(self, filename=None, start_offset=0, compression_type=None, status=None):
        self.event_strs = []
        self.filename = filename
        self.start_offset = start_offset
        self.compression_type = compression_type
        self.status = status
        self.done_reading = False
        self.index = -1

    def read(self):
        if self.filename is None:
            raise errors.NotFoundError(
                None, None, 'No filename provided, cannot read Events')
        if not os.path.exists(self.filename):
            raise errors.NotFoundError(
                None, None,
                '{} does not point to valid Events file'.format(self.filename))

        # TODO: Handle gzip and zlib compressed files
        with open(self.filename, "rb") as f:
            buf = f.read()
            n = self.start_offset

            while n < len(buf):
                # Read the header
                header_str = buf[n:n+8]
                header = struct.unpack('Q', header_str)
                n += 12

                # Read the crc32, which is 4 bytes, and check it against the
                # crc32 of the header
                # crc_header_str = buf[n:n+4]
                # crc_header = struct.unpack('I', crc_header_str)
                # n += 4
                # header_crc_calc = masked_crc32c(header_str)
                # assert header_crc_calc == crc_header[0], \
                #     'Header crc\'s dont match'

                # The length of the header tells us how many bytes the Event
                # string takes
                header_len = int(header[0])
                event_str = buf[n:n+header_len]
                # event_crc_calc = masked_crc32c(event_str)

                n += header_len + 4

                # The next 4 bytes contain the crc32 of the Event string,
                # which we check for integrity. Sometimes, the last Event
                # has no crc32, in which case we skip.
                # if len(buf[n:]) > 0:
                #     crc_event_str = struct.unpack('I', buf[n:n+4])
                #     assert event_crc_calc == crc_event_str[0], \
                #         'Header crc\'s dont match'

                # n += 4
                self.event_strs += [event_str]

            self.done_reading = True

    def GetNext(self):
        if not self.done_reading:
            self.read()

        if self.index >= len(self.event_strs) - 1:
            raise errors.OutOfRangeError(None, None, 'No more events to read')
        else:
            self.index += 1

    def record(self):
        return self.event_strs[self.index]



# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Image summaries and TensorFlow operations to create them, V2 versions.

An image summary stores the width, height, and PNG-encoded data for zero
or more images in a rank-1 string array: `[w, h, png0, png1, ...]`.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorboard.compat import tf2 as tf
from tensorboard.plugins.image import metadata


def image(name,
          data,
          step,
          max_outputs=3,
          description=None):
  """Write an image summary.

  Arguments:
    name: A name for this summary. The summary tag used for TensorBoard will
      be this name prefixed by any active name scopes.
    data: A `Tensor` representing pixel data with shape `[k, h, w, c]`,
      where `k` is the number of images, `h` and `w` are the height and
      width of the images, and `c` is the number of channels, which
      should be 1, 2, 3, or 4 (grayscale, grayscale with alpha, RGB, RGBA).
      Any of the dimensions may be statically unknown (i.e., `None`).
      Floating point data will be clipped to the range [0,1).
    step: Required `int64`-castable monotonic step value.
    max_outputs: Optional `int` or rank-0 integer `Tensor`. At most this
      many images will be emitted at each step. When more than
      `max_outputs` many images are provided, the first `max_outputs` many
      images will be used and the rest silently discarded.
    description: Optional long-form description for this summary, as a
      constant `str`. Markdown is supported. Defaults to empty.

  Returns:
    True on success, or false if no summary was emitted because no default
    summary writer was available.
  """
  summary_metadata = metadata.create_summary_metadata(
      display_name=None, description=description)
  with tf.summary.summary_scope(
      name, 'image_summary', values=[data, max_outputs, step]) as (tag, _):
    tf.debugging.assert_rank(data, 4)
    tf.debugging.assert_non_negative(max_outputs)
    images = tf.image.convert_image_dtype(data, tf.uint8, saturate=True)
    limited_images = images[:max_outputs]
    encoded_images = tf.map_fn(tf.image.encode_png, limited_images,
                               dtype=tf.string,
                               name='encode_each_image')
    # Workaround for map_fn returning float dtype for an empty elems input.
    encoded_images = tf.cond(
        tf.shape(input=encoded_images)[0] > 0,
        lambda: encoded_images, lambda: tf.constant([], tf.string))
    image_shape = tf.shape(input=images)
    dimensions = tf.stack([tf.as_string(image_shape[2], name='width'),
                           tf.as_string(image_shape[1], name='height')],
                          name='dimensions')
    tensor = tf.concat([dimensions, encoded_images], axis=0)
    return tf.summary.write(
        tag=tag, tensor=tensor, step=step, metadata=summary_metadata)



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Internal information about the pr_curves plugin."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorboard.compat.proto import summary_pb2
from tensorboard.plugins.pr_curve import plugin_data_pb2
from tensorboard.util import tb_logging

logger = tb_logging.get_logger()

PLUGIN_NAME = 'pr_curves'

# Indices for obtaining various values from the tensor stored in a summary.
TRUE_POSITIVES_INDEX = 0
FALSE_POSITIVES_INDEX = 1
TRUE_NEGATIVES_INDEX = 2
FALSE_NEGATIVES_INDEX = 3
PRECISION_INDEX = 4
RECALL_INDEX = 5

# The most recent value for the `version` field of the
# `PrCurvePluginData` proto.
PROTO_VERSION = 0

def create_summary_metadata(display_name, description, num_thresholds):
  """Create a `summary_pb2.SummaryMetadata` proto for pr_curves plugin data.

  Arguments:
    display_name: The display name used in TensorBoard.
    description: The description to show in TensorBoard.
    num_thresholds: The number of thresholds to use for PR curves.

  Returns:
    A `summary_pb2.SummaryMetadata` protobuf object.
  """
  pr_curve_plugin_data = plugin_data_pb2.PrCurvePluginData(
      version=PROTO_VERSION, num_thresholds=num_thresholds)
  content = pr_curve_plugin_data.SerializeToString()
  return summary_pb2.SummaryMetadata(
      display_name=display_name,
      summary_description=description,
      plugin_data=summary_pb2.SummaryMetadata.PluginData(
          plugin_name=PLUGIN_NAME,
          content=content))


def parse_plugin_metadata(content):
  """Parse summary metadata to a Python object.

  Arguments:
    content: The `content` field of a `SummaryMetadata` proto
      corresponding to the pr_curves plugin.

  Returns:
    A `PrCurvesPlugin` protobuf object.
  """
  if not isinstance(content, bytes):
    raise TypeError('Content type must be bytes')
  result = plugin_data_pb2.PrCurvePluginData.FromString(content)
  if result.version == 0:
    return result
  else:
    logger.warn(
        'Unknown metadata version: %s. The latest version known to '
        'this build of TensorBoard is %s; perhaps a newer build is '
        'available?', result.version, PROTO_VERSION)
    return result



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""TensorBoard helper routine module.

This module is a trove of succinct generic helper routines that don't
pull in any heavyweight dependencies.

Only pulls in TensorFlow as a dependency if requested per build rules.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import locale
import logging
import os
import re
import sys
import time

from absl import logging as absl_logging
import six

from tensorboard.compat import tf
from tensorboard.util import tb_logging

logger = tb_logging.get_logger()


# TODO(stephanwlee): Move this to program.py
def setup_logging():
  """Configures Python logging the way the TensorBoard team likes it.

  This should be called exactly once at the beginning of main().
  """
  # TODO(stephanwlee): Check the flag passed from CLI and set it to WARN only
  # it was not explicitly set
  absl_logging.set_verbosity(absl_logging.WARN)


def closeable(class_):
  """Makes a class with a close method able to be a context manager.

  This decorator is a great way to avoid having to choose between the
  boilerplate of __enter__ and __exit__ methods, versus the boilerplate
  of using contextlib.closing on every with statement.

  Args:
    class_: The class being decorated.

  Raises:
    ValueError: If class didn't have a close method, or already
        implements __enter__ or __exit__.
  """
  if 'close' not in class_.__dict__:
    # coffee is for closers
    raise ValueError('Class does not define a close() method: %s' % class_)
  if '__enter__' in class_.__dict__ or '__exit__' in class_.__dict__:
    raise ValueError('Class already defines __enter__ or __exit__: ' + class_)
  class_.__enter__ = lambda self: self
  class_.__exit__ = lambda self, t, v, b: self.close() and None
  return class_


def close_all(resources):
  """Safely closes multiple resources.

  The close method on all resources is guaranteed to be called. If
  multiple close methods throw exceptions, then the first will be
  raised and the rest will be logged.

  Args:
    resources: An iterable of object instances whose classes implement
        the close method.

  Raises:
    Exception: To rethrow the last exception raised by a close method.
  """
  exc_info = None
  for resource in resources:
    try:
      resource.close()
    except Exception as e:  # pylint: disable=broad-except
      if exc_info is not None:
        logger.error('Suppressing close(%s) failure: %s', resource, e,
                         exc_info=exc_info)
      exc_info = sys.exc_info()
  if exc_info is not None:
    six.reraise(*exc_info)


def guarded_by(field):
  """Indicates method should be called from within a lock.

  This decorator is purely for documentation purposes. It has the same
  semantics as Java's @GuardedBy annotation.

  Args:
    field: The string name of the lock field, e.g. "_lock".
  """
  del field
  return lambda method: method


class Retrier(object):
  """Helper class for retrying things with exponential back-off."""

  DELAY = 0.1

  def __init__(self, is_transient, max_attempts=8, sleep=time.sleep):
    """Creates new instance.

    :type is_transient: (Exception) -> bool
    :type max_attempts: int
    :type sleep: (float) -> None
    """
    self._is_transient = is_transient
    self._max_attempts = max_attempts
    self._sleep = sleep

  def run(self, callback):
    """Invokes callback, retrying on transient exceptions.

    After the first failure, we wait 100ms, and then double with each
    subsequent failed attempt. The default max attempts is 8 which
    equates to about thirty seconds of sleeping total.

    :type callback: () -> T
    :rtype: T
    """
    failures = 0
    while True:
      try:
        return callback()
      except Exception as e:  # pylint: disable=broad-except
        failures += 1
        if failures == self._max_attempts or not self._is_transient(e):
          raise
        logger.warn('Retrying on transient %s', e)
        self._sleep(2 ** (failures - 1) * Retrier.DELAY)


class LogFormatter(logging.Formatter):
  """Google style log formatter.

  The format is in essence the following:

      [DIWEF]mmdd hh:mm:ss.uuuuuu thread_name file:line] msg

  This class is meant to be used with LogHandler.
  """

  DATE_FORMAT = '%m%d %H:%M:%S'
  LOG_FORMAT = ('%(levelname)s%(asctime)s %(threadName)s '
                '%(filename)s:%(lineno)d] %(message)s')

  LEVEL_NAMES = {
      logging.FATAL: 'F',
      logging.ERROR: 'E',
      logging.WARN: 'W',
      logging.INFO: 'I',
      logging.DEBUG: 'D',
  }

  def __init__(self):
    """Creates new instance."""
    super(LogFormatter, self).__init__(LogFormatter.LOG_FORMAT,
                                       LogFormatter.DATE_FORMAT)

  def format(self, record):
    """Formats the log record.

    :type record: logging.LogRecord
    :rtype: str
    """
    record.levelname = LogFormatter.LEVEL_NAMES[record.levelno]
    return super(LogFormatter, self).format(record)

  def formatTime(self, record, datefmt=None):
    """Return creation time of the specified LogRecord as formatted text.

    This override adds microseconds.

    :type record: logging.LogRecord
    :rtype: str
    """
    return (super(LogFormatter, self).formatTime(record, datefmt) +
            '.%06d' % (record.created * 1e6 % 1e6))

# TODO(stephanwlee): Remove ANSI after removing loader.py
class Ansi(object):
  """ANSI terminal codes container."""

  ESCAPE = '\x1b['
  ESCAPE_PATTERN = re.compile(re.escape(ESCAPE) + r'\??(?:\d+)(?:;\d+)*[mlh]')
  RESET = ESCAPE + '0m'
  BOLD = ESCAPE + '1m'
  FLIP = ESCAPE + '7m'
  RED = ESCAPE + '31m'
  YELLOW = ESCAPE + '33m'
  MAGENTA = ESCAPE + '35m'
  CURSOR_HIDE = ESCAPE + '?25l'
  CURSOR_SHOW = ESCAPE + '?25h'


# TODO(stephanwlee): Remove LogHandler after removing loader.py
class LogHandler(logging.StreamHandler):
  """Log handler that supports ANSI colors and ephemeral records.

  Colors are applied on a line-by-line basis to non-INFO records. The
  goal is to help the user visually distinguish meaningful information,
  even when logging is verbose.

  This handler will also strip ANSI color codes from emitted log
  records automatically when the output stream is not a terminal.

  Ephemeral log records are only emitted to a teletype emulator, only
  display on the final row, and get overwritten as soon as another
  ephemeral record is outputted. Ephemeral records are also sticky. If
  a normal record is written then the previous ephemeral record is
  restored right beneath it. When an ephemeral record with an empty
  message is emitted, then the last ephemeral record turns into a
  normal record and is allowed to spool.

  This class is thread safe.
  """

  EPHEMERAL = '.ephemeral'  # Name suffix for ephemeral loggers.

  COLORS = {
      logging.FATAL: Ansi.BOLD + Ansi.RED,
      logging.ERROR: Ansi.RED,
      logging.WARN: Ansi.YELLOW,
      logging.INFO: '',
      logging.DEBUG: Ansi.MAGENTA,
  }

  def __init__(self, stream, type_='detect'):
    """Creates new instance.

    Args:
      stream: A file-like object.
      type_: If "detect", will call stream.isatty() and perform system
          checks to determine if it's safe to output ANSI terminal
          codes. If type is "ansi" then this forces the use of ANSI
          terminal codes.

    Raises:
      ValueError: If type is not "detect" or "ansi".
    """
    if type_ not in ('detect', 'ansi'):
      raise ValueError('type should be detect or ansi')
    super(LogHandler, self).__init__(stream)
    self._stream = stream
    self._disable_flush = False
    self._is_tty = (type_ == 'ansi' or
                    (hasattr(stream, 'isatty') and
                     stream.isatty() and
                     os.name != 'nt'))
    self._ephemeral = ''

  def emit(self, record):
    """Emits a log record.

    :type record: logging.LogRecord
    """
    self.acquire()
    try:
      is_ephemeral = record.name.endswith(LogHandler.EPHEMERAL)
      color = LogHandler.COLORS.get(record.levelno)
      if is_ephemeral:
        if self._is_tty:
          ephemeral = record.getMessage()
          if ephemeral:
            if color:
              ephemeral = color + ephemeral + Ansi.RESET
            self._clear_line()
            self._stream.write(ephemeral)
          else:
            if self._ephemeral:
              self._stream.write('\n')
          self._ephemeral = ephemeral
      else:
        self._clear_line()
        if self._is_tty and color:
          self._stream.write(color)
        self._disable_flush = True  # prevent double flush
        super(LogHandler, self).emit(record)
        self._disable_flush = False
        if self._is_tty and color:
          self._stream.write(Ansi.RESET)
        if self._ephemeral:
          self._stream.write(self._ephemeral)
      self.flush()
    finally:
      self._disable_flush = False
      self.release()

  def format(self, record):
    """Turns a log record into a string.

    :type record: logging.LogRecord
    :rtype: str
    """
    message = super(LogHandler, self).format(record)
    if not self._is_tty:
      message = Ansi.ESCAPE_PATTERN.sub('', message)
    return message

  def flush(self):
    """Flushes output stream."""
    self.acquire()
    try:
      if not self._disable_flush:
        super(LogHandler, self).flush()
    finally:
      self.release()

  def _clear_line(self):
    if self._is_tty and self._ephemeral:
      # We're counting columns in the terminal, not bytes. So we don't
      # want to take UTF-8 or color codes into consideration.
      text = Ansi.ESCAPE_PATTERN.sub('', tf.compat.as_text(self._ephemeral))
      self._stream.write('\r' + ' ' * len(text) + '\r')
